import ast
import sys
import os
 
def get_annotation(annotation):
	if annotation:
		return ast.unparse(annotation)
	return ""

def extract_info(module, level=0):
	nodes_to_remove = []

	for node in ast.iter_child_nodes(module):
		if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
			print(f"{'Function' if level == 0 else 'Method'}: {node.name}")
			print(f"{'	' * level}Docstring: {ast.get_docstring(node)}")
			for arg in node.args.args:
				print(f"{'	' * level}	Argument: {arg.arg} {get_annotation(arg.annotation)}".rstrip())
			if node.returns:
				print(f"{'	' * level}	Returns: {get_annotation(node.returns)}")

		elif isinstance(node, ast.ClassDef):
			base_classes = ', '.join(ast.unparse(base) for base in node.bases)
			print(f"{'	' * level}Class: {node.name}({base_classes})")
			instance_vars = set()
			for body_node in node.body:
				if isinstance(body_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
					for stmt in body_node.body:
						if isinstance(stmt, ast.Assign):
							for target in stmt.targets:
								if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name) and target.value.id == 'self':
									instance_vars.add(target.attr)
			if instance_vars:
				print(f"{'	' * level}	Instance Variables: {', '.join(sorted(instance_vars))}")
			extract_info(node, level + 1)

		if level == 0:
			nodes_to_remove.append(node)

	for node in nodes_to_remove:
		module.body.remove(node)


def describe_files(files, recursive=False):
	for path in files:
		if os.path.isfile(path) and path.endswith('.py'):
			print(f"File: {os.path.abspath(path)}")
			with open(path, 'r') as f:
				code = f.read()
			module = ast.parse(code)
			extract_info(module)
			print("\n")
		elif os.path.isdir(path):
			if recursive:
				for root, _, file_names in os.walk(path):
					sub_files = [os.path.join(root, name) for name in file_names if name.endswith('.py')]
					describe_files(sub_files, recursive)
			else:
				file_names = [f for f in os.listdir(path) if f.endswith('.py')]
				sub_files = [os.path.join(path, name) for name in file_names]
				describe_files(sub_files, recursive)

if __name__ == '__main__':
	args = sys.argv[1:]
	recursive = '-r' in args
	if recursive:
		args.remove('-r')
	describe_files(args, recursive)



import sys
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QHBoxLayout, QTextEdit, QLineEdit, QPushButton, QScrollArea, QLabel
from ModelProgrammer.Conversation import Conversation, Message
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QFrame, QAbstractItemView
from PyQt5.QtCore import Qt

from PyQt5.QtWidgets import QFrame, QCheckBox, QVBoxLayout, QHBoxLayout, QCompleter, QLabel, QComboBox, QTextEdit, QSizePolicy, QToolButton
from PyQt5.QtCore import Qt, QSize, QPoint, QSettings, QByteArray, pyqtSignal, QRect
from PyQt5.QtGui import QIcon, QPixmap, QImage, QTextOption, QColor, QPalette, QPainter
from typing import Dict, List, Tuple, Optional, Union, Callable, Any

from ModelProgrammer.TokenCounter import tokens_in_message, tokens_in_string

class ColoredFrame(QFrame):
	def __init__(self, background_color, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.background_color = background_color
		self.selected = False

	def paintEvent(self, event):
		if not self.selected:
			painter = QPainter(self)
			painter.fillRect(QRect(0, 0, self.width(), self.height()), self.background_color)
			
		super().paintEvent(event)

	def set_selected(self, selected):
		self.selected = selected
		self.update()
		
class MessageView(ColoredFrame):
	rowHeightChanged = pyqtSignal()
	message_changed = pyqtSignal(Message, str) # Message, old hash
	
	def __init__(self, message: Message, parent=None):
		background_color = QColor( parent.color_palette.get(message.full_role.lower(), QColor(Qt.white)) )
		super().__init__(background_color, parent)
		
		self.parent = parent
		self.message = message
		
		self.layout = QHBoxLayout()
		self.setLayout(self.layout)
		
		self.left_layout = QVBoxLayout()
		self.layout.addLayout(self.left_layout)
		
		# Role (and optional name) label
		self.role_label = QLabel()
		#Get the role from the message, capitalizing the first letter of each word:
		pascal_role = " ".join([word.capitalize() for word in message.full_role.split(" ")])
		self.role_label.setText(f"{pascal_role}:")
		self.role_label.setFixedWidth(100)
		self.role_label.setWordWrap(True)
		self.left_layout.addWidget(self.role_label)
		
		# Date label
		self.date_label = QLabel()
		self.date_label.setText(message.date.strftime("%Y-%m-%d %H:%M:%S"))
		self.date_label.setFixedWidth(100)
		self.date_label.setWordWrap(True)
		self.left_layout.addWidget(self.date_label)
		
		# Token count label
		self.token_count_label = QLabel()
		self.token_count_label.setText(f"{tokens_in_message(message)} tokens")
		self.token_count_label.setFixedWidth(100)
		self.token_count_label.setWordWrap(True)
		self.left_layout.addWidget(self.token_count_label)
		
		# Should send toggle
		self.should_send_checkbox = QCheckBox("Send")
		self.should_send_checkbox.setChecked(message.should_send)
		self.should_send_checkbox.stateChanged.connect(self.on_should_send_changed)
		self.left_layout.addWidget(self.should_send_checkbox)
		
		# Spacer
		self.left_layout.addStretch()
		
		# Editable text box
		self.text_edit = QTextEdit()
		self.text_edit.setPlainText(message['content'])
		self.text_edit.setLineWrapMode(QTextEdit.WidgetWidth)
		self.text_edit.setWordWrapMode(QTextOption.WrapAtWordBoundaryOrAnywhere)
		self.text_edit.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
		self.text_edit.textChanged.connect(self.on_text_changed)
		self.text_edit.textChanged.connect(self.update_text_edit_height)
		self.layout.addWidget(self.text_edit)

		# Vertical panel
		self.panel_layout = QVBoxLayout()
		self.panel_layout.setAlignment(Qt.AlignTop)
		self.layout.addLayout(self.panel_layout)

		# Delete button (X)
		self.delete_btn = QPushButton("X")
		self.delete_btn.clicked.connect(self.delete_message)
		self.delete_btn.setFixedWidth(25)
		self.panel_layout.addWidget(self.delete_btn, alignment=Qt.AlignTop)
		self.delete_btn.clicked.connect(lambda: self.parent.delete_message(self))
		
		# Confirm button (checkmark)
		self.confirm_btn = QPushButton("✓")
		self.confirm_btn.clicked.connect(self.confirm_changes)
		self.confirm_btn.setVisible(False)
		self.confirm_btn.setFixedWidth(25)
		self.panel_layout.addWidget(self.confirm_btn, alignment=Qt.AlignTop)

		# Expand message view button (rotating arrow)
		self.expand_btn = QToolButton()
		self.expand_btn.setCheckable(True)
		self.expand_btn.setArrowType(Qt.RightArrow)
		self.expand_btn.toggled.connect(self.toggle_expand)
		self.panel_layout.addWidget(self.expand_btn, alignment=Qt.AlignTop)

		self.update_text_edit_height()
	
	def on_should_send_changed(self, state):
		self.message.should_send = state == Qt.Checked
		# self.parent.message_changed.emit(self.message)
	
	def update_text_edit_height(self):
		new_height = int(self.delete_btn.sizeHint().height() * 3)
		margins = self.text_edit.contentsMargins()
		if self.expand_btn.arrowType() == Qt.DownArrow:
			new_height = max(int(self.text_edit.document().size().height()), new_height) + margins.top() + margins.bottom()
			self.text_edit.setFixedHeight(new_height)
		else:
			new_height = int(self.delete_btn.sizeHint().height() * 3) + margins.top() + margins.bottom()
			self.text_edit.setFixedHeight(new_height)
		self.rowHeightChanged.emit()
		
	def toggle_expand(self, checked):
		self.expand_btn.setArrowType(Qt.DownArrow if checked else Qt.RightArrow)
		self.update_text_edit_height()
		
	def on_text_changed(self):
		text = self.text_edit.toPlainText()
		if text != self.message['content']:
			self.confirm_btn.setVisible(True)
			# self.text_edit.setStyleSheet("border: 3px solid rgba(0, 0, 255, 0.3);")
			self.text_edit.setStyleSheet("QTextEdit {border: 3px solid rgba(0, 0, 255, 0.3);}")

		else:
			self.confirm_btn.setVisible(False)
			self.text_edit.setStyleSheet("")
			
		if self.expand_btn.isChecked():
			self.text_edit.setMinimumHeight(int(self.text_edit.document().size().height()))
			self.text_edit.updateGeometry()
			self.rowHeightChanged.emit()
		
		self.token_count_label.setText(f"{tokens_in_string(text)} tokens")
		
	def delete_message(self):
		pass

	def confirm_changes(self):
		old_hash = self.message.hash
		self.message.json['content'] = self.text_edit.toPlainText()
		self.confirm_btn.setVisible(False)
		self.text_edit.setStyleSheet("")
		self.message.recompute_hash()
		
		self.message_changed.emit(self.message, old_hash)

class ConversationView(QListWidget):
	message_changed = pyqtSignal(Message, str) # Message, old hash
	
	def __init__(self, conversation: Conversation, *args, **kwargs):
		super().__init__(*args, **kwargs)
		
		self.colors = QColor.colorNames()#  ['red', 'green', 'blue', 'yellow', 'cyan', 'magenta', 'gray']
		
		self._color_palette = {
			"system": "lightgrey",
			"user human": "#9DFFA6",
			"user terminal": "#FFEBE4",
			"assistant": "#FFC4B0",
		}
		
		self.conversation = conversation
		
		for message in conversation.messages:
			self.render_message(message)
			
		self.setAutoScroll(False)
		self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
		self.setSelectionMode(QAbstractItemView.ExtendedSelection)
		self.itemSelectionChanged.connect(self.update_selection)
	
	@property
	def color_palette(self):
		if not hasattr(self, '_color_palette') or self._color_palette is None:
			self._color_palette = {}
			message_types = set()

			# Extract message types from the messages in the conversation
			for message in self.conversation.messages:
				message_types.add(message.full_role.lower())

			# Assign colors to message types
			for i, message_type in enumerate(sorted(message_types)):
				message_type = message_type.lower()
				self._color_palette[message_type] = self.colors[i % len(self.colors)]

		return self._color_palette
		
	def update_selection(self):
		for index in range(self.count()):
			item = self.item(index)
			message_view = self.itemWidget(item)
			if message_view is not None:
				message_view.set_selected(item.isSelected())
	
	def keyPressEvent(self, event):
		item_widget = self.itemWidget(self.currentItem())
		if event.key() == Qt.Key_Delete and not item_widget.text_edit.hasFocus():
			#if selection is not empty, delete all selected messages
			if self.selectedItems():
				for item in self.selectedItems():
					item_widget = self.itemWidget(item)
					self.delete_message(item_widget)
		elif event.key() == Qt.Key_Escape:
			self.clearSelection()
		else:
			super().keyPressEvent(event)
	
	def delete_message(self, message_widget): #TODO: this needs to take a message not a widget
		if message_widget is not None:
			item = None
			for index in range(self.count()):
				current_item = self.item(index)
				if self.itemWidget(current_item) == message_widget:
					item = current_item
					break

			if item is not None:
				row = self.row(item)
				item = self.takeItem(row)

				message_to_remove = message_widget.message
				self.conversation.messages.remove(message_to_remove)
				
				self.clearSelection()
	
	def render_message(self, message: Message):
		message_type = message.full_role.lower()
		if message_type not in self.color_palette:
			self._color_palette[message_type] = self.colors[len(self._color_palette) % len(self.colors)]
			
		item = QListWidgetItem()
		item_widget = MessageView(message, self)
		item_widget.rowHeightChanged.connect(lambda: self.update_row_height(item))
		def message_changed(message: Message, old_hash: str):
			self.message_changed.emit(message, old_hash)
			self.conversation.recompute_hash()
			self.update_row_height(item)
		item_widget.message_changed.connect(message_changed)
		item.setSizeHint(item_widget.sizeHint())
		self.addItem(item)
		self.setItemWidget(item, item_widget)
		self.scrollToBottom()
				
	def add_message(self, message: Message):
		self.conversation.add_message(message)
		self.render_message(message)
		
	def update_row_height(self, item: QListWidgetItem):
		item_widget = self.itemWidget(item)
		item.setSizeHint(QSize(item_widget.sizeHint().width(), item_widget.sizeHint().height()))

class RoleComboBox(QComboBox):
	def __init__(self, values:List[str], default_value:str="Human", *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.setEditable(True)
		self.default_value = default_value
		self.addItems(values)
		self.set_default_value()
		#do not auto complete
		self.completer().setCompletionMode(QCompleter.UnfilteredPopupCompletion)
		self.setInsertPolicy(QComboBox.NoInsert)
		
	def set_default_value(self):
		index = self.findText(self.default_value)
		if index != -1:
			self.setCurrentIndex(index)
			
class ChatUI(QWidget):
	#qt signal for sending messages
	message_added = pyqtSignal(Conversation, Message, bool)
	message_changed = pyqtSignal(Message, str) # Message, old hash
	confirm_command = pyqtSignal()
	
	def __init__(self, conversation: Conversation, roles:List[str], max_new_message_lines=5):
		super().__init__()
		
		self.conversation = conversation
		self.roles = roles
		
		self.max_new_message_lines = max_new_message_lines
		self.num_lines = 0
		
		self.init_ui()
		self.has_unrun_command = False
		
		self.read_settings()
		
	def init_ui(self):
		self.setWindowTitle('Chat')

		self.layout = QVBoxLayout()
		self.setLayout(self.layout)

		self.list_view = ConversationView(self.conversation)
		self.list_view.message_changed.connect(self.message_changed.emit)
		self.layout.addWidget(self.list_view)
		
		self.input_layout = QHBoxLayout()
		self.role_combobox = RoleComboBox(self.roles, default_value="Human")
		self.input_layout.addWidget(self.role_combobox, alignment=Qt.AlignBottom)

		self.input_field = QTextEdit()
		size_policy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
		size_policy.setVerticalStretch(1)
		self.input_field.setSizePolicy(size_policy)
		self.input_field.setMaximumHeight(self.input_field.fontMetrics().lineSpacing())
		self.input_field.setWordWrapMode(QTextOption.WrapAtWordBoundaryOrAnywhere)
		self.input_field.textChanged.connect(self.adjust_input_field_size)
		self.input_field.textChanged.connect(self.update_send_button_text)
		self.input_field.setPlaceholderText("Type your message here...")
		self.input_layout.addWidget(self.input_field, alignment=Qt.AlignBottom)
		
		self.send_button = QPushButton('Send')
		self.send_button.clicked.connect(self.send_message)
		
		self.send_add_toggle = QCheckBox()
		self.send_add_toggle.setCheckState(Qt.Checked)
		self.send_add_toggle.stateChanged.connect(lambda: self.update_send_button_text())
		self.input_layout.addWidget(self.send_add_toggle, alignment=Qt.AlignBottom)
		
		self.input_layout.addWidget(self.send_button, alignment=Qt.AlignBottom)
		
		self.layout.addLayout(self.input_layout)
		self.input_field.setMinimumHeight(self.send_button.sizeHint().height())
		
	def read_settings(self):
		settings = QSettings("MyCompany", "MyApp")
		self.restoreGeometry(settings.value("geometry", QByteArray()))

	def write_settings(self):
		settings = QSettings("MyCompany", "MyApp")
		settings.setValue("geometry", self.saveGeometry())
		
	def send_message(self):
		message = None
		
		message_text = self.input_field.toPlainText()
		message = Message.from_role_content(self.role_combobox.currentText(), message_text)
			
		self.message_added.emit(
			self.conversation, 
			message,
			self.send_add_toggle.isChecked()
		)
		
		self.input_field.clear()
		
	def closeEvent(self, event):
		self.write_settings()
		super().closeEvent(event)
		
	def keyPressEvent(self, event):
		if event.key() == Qt.Key_Enter and event.modifiers() == Qt.ControlModifier:
			self.send_button.click()
		else:
			super().keyPressEvent(event)
	
	def adjust_input_field_size(self):
		"""Adjust the height of the input field to fit the text up to max lines"""
		n_lines = self.input_field.document().blockCount()
		lines_to_show = min(n_lines, self.max_new_message_lines)
		new_height = self.input_field.fontMetrics().lineSpacing() * lines_to_show + 10
		self.input_field.setMaximumHeight(int(new_height))

		if n_lines >= self.max_new_message_lines:
			self.input_field.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
		else:
			self.input_field.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
			
		self.input_field.updateGeometry()
		
		if self.num_lines < n_lines:
			self.input_field.verticalScrollBar().setValue(self.input_field.verticalScrollBar().maximum())
		self.num_lines = n_lines
	
	def render_message(self, message:Message):
		self.list_view.render_message(message)
	
	def update_send_button_text(self):
		#"Run" if has unrun command, change to run, and the input field is empty:
		if self.has_unrun_command and self.input_field.toPlainText() == "":
			if self.send_button.text() != "Run":
				self.send_button.setText("Run")
				try:
					self.send_button.clicked.disconnect(self.send_message)
				except TypeError:
					pass
				self.send_button.clicked.connect(self.confirm_command.emit)
		else:
			if self.send_button.text() == "Run":
				try:
					self.send_button.clicked.disconnect(self.confirm_command.emit)
				except TypeError:
					pass
				self.send_button.clicked.connect(self.send_message)
			self.send_button.setText("Send" if self.send_add_toggle.isChecked() else "Add")
	
	@property
	def has_unrun_command(self):
		return self._has_unrun_command
	
	@has_unrun_command.setter
	def has_unrun_command(self, value:bool):
		self._has_unrun_command = value
		self.update_send_button_text()


import asyncio
import os
import pty
import signal
import subprocess
import re

class Terminal:
	def __init__(self):
		self.master, self.slave = pty.openpty()
		self.process = subprocess.Popen(
			["bash"],
			stdin=self.slave,
			stdout=self.slave,
			stderr=self.slave,
			bufsize=0,
			preexec_fn=os.setsid,
		)
		
	async def _read_output(self):
		"""Clean the output from the terminal."""
		loop = asyncio.get_event_loop()
		output = await loop.run_in_executor(None, os.read, self.master, 1024)
		decoded_output = output.decode()
		ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
		cleaned_output = ansi_escape.sub('', decoded_output)
		extra_characters = re.compile(r'(\x07|^0;[^$]*)(?=.*?@)')
		return extra_characters.sub('', cleaned_output)
		
	async def run_command(self, command:str, timeout:float=None)->str:
		"""Run the passed command and return the output."""
		
		os.write(self.master, f"{command}\n".encode())

		lines = []
		while True:
			try:
				line = await asyncio.wait_for(self._read_output(), timeout)
				if line:
					lines.append(line.rstrip())
				else:
					break
			except asyncio.TimeoutError:
				print("Command timed out")
				await self.stop()
				break

		return "\n".join(lines)

	async def stop(self):
		"""Stop any running command"""
		os.killpg(os.getpgid(self.process.pid), signal.SIGINT)


import openai
import os

#Set the OpenAI API key path to ~/.config/ModelProgrammer/API_key.txt, expanding the ~ to the user's home directory
openai.api_key_path = os.path.expanduser("~/.config/ModelProgrammer/API_key.txt")

#print(f"Models available:{openai.Model.list()}")


from ModelProgrammer.Conversation import Conversation, Message
from ModelProgrammer.Terminal import Terminal
from typing import List, Dict
import tiktoken
import asyncio
import json

import openai
import datetime

class ChatBot:
	def __init__(self, model="gpt-3.5-turbo", temperature=0):
		self.terminal = Terminal()
		self.model = model
		self.temperature = temperature
		
	async def handle_response(self, chat_bot_response: str) -> str:
		response_data = json.loads(chat_bot_response)
		message_content = response_data["choices"][0]["message"]["content"]
		command = message_content.replace(r'\n', '\n')
		terminal_output = await self.terminal.run_command(command)
		return terminal_output
	
	def count_conversation_tokens(self, conversation:Conversation) -> int:
		def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0301"):
			"""Returns the number of tokens used by a list of messages."""
			try:
				encoding = tiktoken.encoding_for_model(self.model)
			except KeyError:
				encoding = tiktoken.get_encoding("cl100k_base")
			if model == "gpt-3.5-turbo-0301":  # note: future models may deviate from this
				num_tokens = 0
				for message in messages:
					num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
					for key, value in message.items():
						num_tokens += len(encoding.encode(value))
						if key == "name":  # if there's a name, the role is omitted
							num_tokens += -1  # role is always required and always 1 token
				num_tokens += 2  # every reply is primed with <im_start>assistant
				return num_tokens
			else:
				raise NotImplementedError(f"""num_tokens_from_messages() is not presently implemented for model {model}.
		See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens.""")
		return num_tokens_from_messages(conversation.as_list())
		
	def count_message_tokens(self, message: str) -> int:
		encoding = tiktoken.encoding_for_model(self.model)
		tokens = encoding.encode(message)
		return len(tokens)
		
	async def send_conversation(self, conversation:Conversation, max_tokens:int=150) -> Dict[str,str]:
		conversation_tokens = self.count_conversation_tokens(conversation)
		max_tokens = min(4096 - conversation_tokens, max_tokens)
		if max_tokens < 0: #TODO: get this limit per model type
			raise ValueError(f"Conversation is too long for model {self.model} at {conversation_tokens} tokens, only 4096 are supported.")
		
		#Send the messages to OpenAI's Chat API
		pre_response_time = datetime.datetime.now()
		response = openai.ChatCompletion.create(
			model=self.model,
			messages=conversation.as_list(),
			temperature=self.temperature,
			max_tokens=max_tokens,
		)
		
		#print how long it's been in seconds since the request was sent, with a helpful message
		print(f"OpenAI's Chat API took {datetime.datetime.now() - pre_response_time} seconds to respond.")
		date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
		
		return response


import sqlite3
import hashlib
import json

import datetime
from typing import List, Dict

class Hashable():
	def __init__(self):
		self.hash = ""
	
	def recompute_hash(self):
		raise NotImplementedError()
		
	def _compute_hash(self, x:object) -> str:
		return hashlib.sha256(str(x).encode("utf-8")).hexdigest()
		
class Message(Hashable):
	def __init__(self, data:Dict[str, str], date=None):
		self.json = data
		self.should_send = True
		if date is None:
			if "created" in data:
				self.date = datetime.datetime.fromtimestamp(int(data["created"]))
			else:
				self.date = datetime.datetime.now()
		self.recompute_hash()
		
	def recompute_hash(self) -> str:
		self.hash = self._compute_hash(json.dumps(self.json))
	
	@classmethod
	def from_role_content(cls, role:str, content:str):
		role = role.lower()
		if role in ["user", "assistant", "system"]:
			return cls({"role":role, "content":content})
		else:
			return cls({"role":"user", "name":role, "content":content})
	
	@property
	def full_role(self):
		if "role" not in self.json:
			return None
		role = self.json['role']
		if 'name' in self.json:
			return f"{role} {self.json['name']}".strip()
		return role.strip()
	
	@property
	def short_role(self):
		if 'name' in self.json:
			return self.json['name']
		else:
			return self.json['role']
	
	@property
	def content(self):
		if 'content' in self.json:
			return self.json['content']
		else:
			if 'choices' in self.json:
				return self.json['choices'][0]['message']['content']
		return None
	@content.setter
	def content(self, value):
		if 'content' in self.json:
			self.json['content'] = value
		else:
			if 'choices' in self.json:
				self.json['choices'][0]['message']['content'] = value
		
	def __str__(self):
		return json.dumps(self.json)
		
	def __getitem__(self, key):
		return self.json.get(key, None)

class Conversation(Hashable):
	def __init__(self, messages:List[Message]):
		self.messages = messages
		self.hash = ""
		self.recompute_hash()
	
	@property
	def hashes(self):
		"""
		Get the hashes of all messages in the conversation.
		"""
		return [message.hash for message in self.messages]
		
	def recompute_hash(self):
		self.hash = self._compute_hash(self.hashes)
	
	@classmethod
	def from_list(cls, messages:List[Dict[str, str]]):
		return cls([Message(message) for message in messages])
		
	def as_list(self, include_all:bool=False):
		"""
		Generates a sendable list of messages json's ready to be sent to OpenAI's Chat API.
		
		Args:
			include_all (bool): If True, all messages will be included, even if they are not marked as "should_send".
			
		Returns:
			list: A list of message dictionaries, each of which has a "role" key and a "content" key, some also have a "name" key.
		"""
		return [message.json for message in self.messages if message.should_send or include_all]
	
	def add_message(self, message:Message):
		self.messages.append(message)
		self.recompute_hash()
		
	def __str__(self):
		return json.dumps([str(message) for message in self.messages])


from ModelProgrammer.ChatUI import *
from ModelProgrammer.Conversation import *
from ModelProgrammer.ConversationDB import *
from ModelProgrammer.ChatBot import *

class Programmer():
	def __init__(self, db:ConversationDB, conversation:Conversation=None, terminal:Terminal=None):
		if conversation is None:
			conversation = Conversation([])
		if terminal is None:
			terminal = Terminal()
		
		self.db = db
		self.conversation = conversation
		self.terminal = terminal
		
		self.programmer = ChatBot()
		
		self.chat_ui = ChatUI(conversation, [
			"Assistant",
			"System",
			"User",
			"Terminal",
			"Human"
		])
		self.chat_ui.message_changed.connect(self.message_changed)
		self.chat_ui.message_added.connect(self.message_added)
		self.chat_ui.confirm_command.connect(self.run_terminal_command)
		self.chat_ui.show()
	
	def message_changed(self, message:Message, old_hash:str):
		self.db.save_message(message, MessageType.Edit, old_hash)
	
	def message_added(self, conversation:Conversation, message:Message, should_send:bool):		
		if message is not None and len(message.content)>0:
			self.db.save_message(message, MessageType.ManualEntry)
			self.chat_ui.render_message(message)
			if len(conversation.messages)>0:
				self.db.save_conversation(conversation)
				
			conversation.add_message(message)
		
		self.db.save_conversation(conversation)
		
		response = None
		if should_send:
			if message is None or message.full_role != "assistant":
				response_json = None
				try:
					response_json = asyncio.run(self.programmer.send_conversation(conversation))
				except Exception as e:
					print(e)
					
				if response_json is not None:
					response = Message(response_json)
					self.db.save_response_for_conversation(conversation, response) #Save the raw response from the chatbot
					response = self.format_response(response) #Format the response as a Message supported by the ChatUI and ChatBot
					self.db.save_message(response, MessageType.FormattedChatbotResponse, response.hash) #Save the formatted response referring to the raw response
					
					self.chat_ui.render_message(response)
					conversation.add_message(response)
					self.db.save_conversation(conversation)
					self.chat_ui.has_unrun_command = True #TODO: check if the response contains a command
		elif message is not None and message.full_role == "assistant":
			response = message
			self.chat_ui.has_unrun_command = True #TODO: check if the response contains a command
		
		#TODO: parse out the command(s) from the response and run them
		self.response = response
	
	def run_terminal_command(self):
		if self.response is None or self.chat_ui.has_unrun_command == False:
			return
		self.chat_ui.has_unrun_command = False
		terminal_output = asyncio.run(self.terminal.run_command(self.response.content, 2))
		output_message = Message.from_role_content("terminal", terminal_output)
		self.db.save_message(output_message, MessageType.TerminalOutput)
		self.conversation.add_message(output_message)
		self.db.save_conversation(self.conversation)
		self.chat_ui.render_message(output_message)
		self.chat_ui.update_send_button_text()
		
	def format_response(self, response:Message) -> Message:
		if response is None:
			return None
		content = response.json["choices"][0]["message"]["content"]
		return Message.from_role_content("assistant", content)


from ModelProgrammer.Conversation import *
from enum import Enum
import datetime
import sqlite3
import hashlib
import json

class MessageType(Enum):
	RawChatbotResponse = 0
	FormattedChatbotResponse = 1
	ManualEntry = 2
	TerminalOutput = 3
	Edit = 4
	Unknown = 5

	def __int__(self):
		return self.value
	
class ConversationDB():
	"""
	Stores conversations in a sqlite database.
	
	Hashes each message and saves it uniquely to a sqlite database by its hash.
	Records in a separate table the list of message hashes that make up each conversation.
	Saves responses to the database by its hash, and records the hash in the conversation table.
	
	Note that all hashes are SHA-256 hashes.
	"""
	def __init__(self, path:str="Chatter.db"):
		self.connection = sqlite3.connect(path)
		self.cursor = self.connection.cursor()
		self.cursor.execute("CREATE TABLE IF NOT EXISTS messages (hash text, json text, PRIMARY KEY (hash))")
		self.cursor.execute("CREATE TABLE IF NOT EXISTS message_info (hash text, datetime text, type integer, role text, source_hash text, PRIMARY KEY (hash))")
		self.cursor.execute("CREATE TABLE IF NOT EXISTS conversations (hash text, datetime text, message_hashes text, PRIMARY KEY (hash))")
		self.connection.commit()
	
	def now(self) -> str:
		'''Get the datetime in 2023-01-01 00:00:00 format'''
		return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
		
	def save_message(self, message:Message, message_type:MessageType, source_hash:str=None):
		"""
		Saves a message to the database.
		"""
		date = self.now()
		
		#Save the message to the database preserving the original message if it exists
		self.cursor.execute("SELECT * FROM messages WHERE hash = ?", (message.hash,))
		if not self.cursor.fetchone():
			self.cursor.execute("INSERT INTO messages VALUES (?, ?)", (message.hash, str(message)))
		
		#Save the message_info to the database preserving the original info if it exists
		self.cursor.execute("SELECT * FROM message_info WHERE hash = ?", (message.hash,))
		if not self.cursor.fetchone():
			self.cursor.execute("INSERT INTO message_info VALUES (?, ?, ?, ?, ?)", (message.hash, date, int(message_type), message.full_role, source_hash))
		self.connection.commit()
				
	def save_conversation(self, conversation:Conversation):
		"""
		Saves a conversation to the database.
		"""
		date = self.now()
		
		#Save the messages to the database
		for message in conversation.messages:
			self.save_message(message, MessageType.Unknown, conversation.hash)
		
		#Save the conversation to the database if it doesn't already exist
		self.cursor.execute("SELECT * FROM conversations WHERE hash = ?", (conversation.hash,))
		if not self.cursor.fetchone():
			self.cursor.execute("INSERT INTO conversations VALUES (?, ?, ?)", (conversation.hash, date, json.dumps(conversation.hashes)))
		
		#Commit the changes to the database
		self.connection.commit()
	
	# save a response from a chatbot to the database for the given conversation
	def save_response_for_conversation(self, conversation:Conversation, response:Message):
		"""
		Saves a response to the database for the given conversation.
		"""
		date = self.now()
		
		#Save the response to the database
		self.cursor.execute("INSERT INTO messages VALUES (?, ?)", (response.hash, json.dumps(response.json)))
		self.connection.commit()
		
		#Save the message_info to the database preserving the original info if it exists
		self.cursor.execute("SELECT * FROM message_info WHERE hash = ?", (response.hash,))
		if not self.cursor.fetchone():
			self.cursor.execute("INSERT INTO message_info VALUES (?, ?, ?, ?, ?)", (response.hash, date, int(MessageType.RawChatbotResponse), "assistant", conversation.hash))
		self.connection.commit()
		
		#Commit the changes to the database
		self.connection.commit()

	@property
	def latest_conversation(self) -> str:
		self.cursor.execute("SELECT hash FROM conversations order by datetime desc limit 1")
		row = self.cursor.fetchone()
		if row is None:
			return None
		return row[0]
	
	def load_message(self, message_hash:str) -> Message:
		self.cursor.execute("SELECT json FROM messages WHERE hash = ?", (message_hash,))
		message_text = self.cursor.fetchone()
		if message_text is None:
			return None
		return Message(json.loads(message_text[0]))
		
	def load_conversation(self, conversation_hash:str) -> Conversation:
		if conversation_hash is None:
			return None
		self.cursor.execute("SELECT message_hashes FROM conversations WHERE hash = ?", (conversation_hash,))
		message_hashes = self.cursor.fetchone()
		if message_hashes is None:
			return None
		return Conversation([self.load_message(message_hash) for message_hash in json.loads(message_hashes[0]) if message_hash is not None])


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ModelProgrammer",
    version="0.0.3",
    author="Charlie Angela Mehlenbeck",
    author_email="charlie_inventor2003@yahoo.com",
    description="An experiment in AI terminal interaction.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/inventor2525/ModelProgrammer",
    packages=setuptools.find_packages(),
    python_requires=">=3.10",
    install_requires=[
        "tiktoken",
        "PyQt5",
        "openai",
    ],
)



import tiktoken
from ModelProgrammer.Conversation import Message, Conversation

def tokens_in_string(s:str) -> int:
	"""Note: This is not exact. See code in ChatBot.py for more details."""
	
	encoding = tiktoken.encoding_for_model("gpt-3.5-turbo")
	tokens = encoding.encode(s)
	return len(tokens)
	
def tokens_in_message(message:Message) -> int:
	"""Note: This is not exact. See code in ChatBot.py for more details."""
	
	return tokens_in_string(message.content)